------------------------------------------------------------------------------------------------------------
--
-- Copyright (C) 2010 Nikolaos Ch. Alachiotis
--
-- Engineer: 			Nikolaos Alachiotis 
-- Contact:				n.alachiotis@gmail.com
-- 						http://wwwkramer.in.tum.de/exelixis/nikos
--
-- Create Date:    	04/06/2010 			 
-- Design Name: 	 	SP_EAU
-- Module Name:    	SP_EAU - Behavioral 
-- Project Name:   	Single-Precision Pipelined Exponential Approximation Unit for FPGAs
-- Target Devices: 	Virtex 5 family
-- Tool versions: 	Xilinx ISE Design Suite 10.1
--
-- Description: 		CORDIC + Table Driven based implementation.
--					 		Original C-code: http://people.sc.fsu.edu/~burkardt/cpp_src/cordic/cordic.html
--					 		Simplified C-based model of this unit is included in the standard release of
--					 		RAxML (R-andomized Ax-elerated M-aximum L-ikelihood): http://wwwkramer.in.tum.de/exelixis/
--							A detailed description and evaluation of the EAU architecture can be found at
--							http://wwwkramer.in.tum.de/exelixis/nikos/publications.html
--
-- Evaluation:			Maximum Clock Frequency: 305 MHz
--							Latency in clock cycles: 194
--							Throughput:					 1 Result/Cycle
--							Accuracy: 					 0.83 * 2^(-17) (Max Error)
--															 0.66 * 2^(-19) (Avg Error)
--
-- Additional Info:	The unit has been verified on Virtex5 SX95T-1 FPGA.
--							Use the "Area Reduction" optimization strategy for slightly better performance.  	
--
------------------------------------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

---- Uncomment the following library declaration if instantiating
---- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

entity SP_EAU is
    Port ( rst : in  STD_LOGIC;  									-- Reset Signal (Active High)
           clk : in  STD_LOGIC;  									-- Clock Signal 
           valid_in : in  STD_LOGIC; 								-- Valid Input Value
           value_in : in  STD_LOGIC_VECTOR (31 downto 0);	-- Input Value
           valid_out : out  STD_LOGIC;								-- Valid Output Value
           value_out : out  STD_LOGIC_VECTOR (31 downto 0));-- Exp(Input Value) 
end SP_EAU;

architecture Behavioral of SP_EAU is

component EAU_PRE_ITER_c is
    Port ( rst : in  STD_LOGIC;
           clk : in  STD_LOGIC;
           vld_in : in  STD_LOGIC;
           fp_val_in : in  STD_LOGIC_VECTOR (31 downto 0);
			  int32_vld_out: out STD_LOGIC;
           int32_val_out : out  STD_LOGIC_VECTOR (31 downto 0);
			  fp_sub_vld_out: out STD_LOGIC;
           fp_sub_val_out : out  STD_LOGIC_VECTOR (31 downto 0));
end component;

signal int32_vld_out, fp_sub_vld_out: std_logic;
signal int32_val_out, fp_sub_val_out: std_logic_vector(31 downto 0);

component EAU_ITER_TOP_c is
    Port ( rst : in  STD_LOGIC;
           clk : in  STD_LOGIC;
           valid_in : in  STD_LOGIC;
           value_in : in  STD_LOGIC_VECTOR (31 downto 0);
           valid_out : out  STD_LOGIC;
           vec_out : out  STD_LOGIC_VECTOR (11 downto 0);
           val_out : out  STD_LOGIC_VECTOR (31 downto 0));
end component;

signal iter_vld_out: std_logic;
signal lut_index: std_logic_vector(11 downto 0);
signal iter_val_out: std_logic_vector(31 downto 0);

component AFTITERPHASE is
    Port ( rst : in  STD_LOGIC;
           clk : in  STD_LOGIC;
           valid_in : in  STD_LOGIC;
           pLUTaddr : in  STD_LOGIC_VECTOR (7 downto 0);
           mLUTaddr : in  STD_LOGIC_VECTOR (11 downto 0);
           iter_val : in  STD_LOGIC_VECTOR (31 downto 0);
           valid_out : out  STD_LOGIC;
           value_out : out  STD_LOGIC_VECTOR (31 downto 0));
end component;

signal pLUTaddr_sftd: std_logic_Vector(7 downto 0);

component shift_reg_8b_149c is
  port (
    clk : in STD_LOGIC := 'X'; 
    d : in STD_LOGIC_VECTOR ( 7 downto 0 ); 
    q : out STD_LOGIC_VECTOR ( 7 downto 0 ) 
  );
end component;

begin

PREITERPHASE: EAU_PRE_ITER_c Port Map
( rst => rst,
  clk => clk,
  vld_in => valid_in,
  fp_val_in => value_in,
  int32_vld_out => int32_vld_out,
  int32_val_out => int32_val_out,
  fp_sub_vld_out => fp_sub_vld_out,
  fp_sub_val_out => fp_sub_val_out
);

ITERLOOP: EAU_ITER_TOP_c Port Map
( rst => rst,
  clk => clk,
  valid_in => fp_sub_vld_out,
  value_in => fp_sub_val_out,
  valid_out => iter_vld_out,
  vec_out => lut_index,
  val_out => iter_val_out
);

SHIFTER1: shift_reg_8b_149c Port Map
(
    clk => clk, 
    d => int32_val_out(7 downto 0), 
    q => pLUTaddr_sftd 
);

FINALPHASE: AFTITERPHASE Port Map
( rst => rst,
  clk => clk,
  valid_in => iter_vld_out,
  pLUTaddr => pLUTaddr_sftd,
  mLUTaddr => lut_index,
  iter_val => iter_val_out,
  valid_out => valid_out,
  value_out => value_out
);

end Behavioral;

